#
#
# $Id: DBCleanStructure.pm,v 1.1.2.13 2012/06/04 10:56:47 kyri Exp $
#
# (C) Copyright 2012 by Open iT inc., All rights reserved.
#
# This software may be used or copied only with the written permission by
# Open Systems Consultans a.s or in accordance with the terms and conditions
# stipulated in the agreement under which this software has been supplied.
#
# This file is part of Open iT.
#
#HEAD##########################################################################
#
# @NAME:
#   DBCleanStructure
# @AUTHOR:
#   Kjetil Yri
# @CREATED:
#   Thu Apr 12 09:55:28 MET 2012
# @DESCRIPTION:
#   Module for file operations.
#
#/HEAD#########################################################################

# Assumptions & methods:
# * All data-*-*-*.temp are considered to be put there by check_and_split_data,
#   and therefore are OK
# * For propagating data:
#   - Only the highest resolution directories are considered for cleanup,
#     assuming that problems at lower resolution would be caused by problems
#     at the higher resolution
#   - A directory is considered for pruning if there is more than one file in
#     the directory, or if there is one file and the file present is not called
#     data or data.gz
#   - If data-*-*-*.temp exist, use the newest and remove all other files,
#     if not, use the oldest data* file
#   - Make sure that the file left after pruning is called data-*-*-*.temp
#     to trigger propagation next time add runs
# * For addition data:
#   - Leave the oldest non-zero of
#     data(.gz) and data(.gz).cpy (usually) / data(.gz).temp (after --redo)
#   - Rename data(.gz).cpy or data(.gz).temp to data(.gz)
#   - Delete other files, except data-*-*-*.temp
# * For links in the W directory
#   - Assume that there has been cleanup on the M directory first
#   - Assume that if there is a day directory pointed to by a link
#     we keep it even if there is no data in it, since the data should
#     have been removed by the M processing

use strict;

package OpeniT::Contrib::DBCleanStructure;

use OpeniT::Contrib::Config;
use OpeniT::Contrib::File;

use File::Basename;
use File::Copy;

my $config = new OpeniT::Contrib::Config;
$config->read( "/etc/opt/openit/openit.cfg" );


require Exporter;

@OpeniT::Contrib::DatabaseExtraction::ISA    = qw( Exporter );
@OpeniT::Contrib::DatabaseExtraction::EXPORT =
  qw( set_info process_data );
@OpeniT::Contrib::DatabaseExtraction::EXPORT_OK = qw();

my $filesep = OpeniT::Contrib::File::file_separator();

my $datatype;
my $is_propagate;
my $resolution;
my $db_path = $config->{DATA_DIR};

my $info = 0;

sub REMOVE_DIR {
    return 1;
}

sub KEEP_DIR {
    return 0;
}


sub set_info {
    $info = shift @_;
}

sub check_propagate {
    my $file = $config->{CONFIG_DIR} . "${filesep}data_generation_config";
    if( ! open( DATA_GEN_CONF, "<$file" ) ) {
	die "Could not open '$file': $!\n";
    }
    my $line;
    while( $line = <DATA_GEN_CONF> ) {
	if( $line =~ m/^$datatype:/ ) {
	    close( DATA_GEN_CONF );
	    &info( "Datatype $datatype is propagate type\n" );
	    $is_propagate = 1;
	    return;
	}
    }
    if( defined $line ) {
	die "Could not read '$file': $!\n";
    }
    close( DATA_GEN_CONF );
    &info( "Datatype $datatype is non-propagate type\n" );
    $is_propagate = 0;
}

sub find_resolution {
    my $dir_name = shift @_;

    # 60/300/900/1800|3600 -> use this if found
    # (month).[123]?\d/data(.gz)? with no subdir -> use 'day' if found
    # If none of the above is found, use 'week'

    if( $dir_name =~ m/$db_path.$datatype.M$/ ) {
	&info( "Finding resolution for data type $datatype\n" );
	$resolution = undef;
    }

    if( defined $resolution ) {
	&info( "Resolution is $resolution\n" );
	return;
    }

    &info( "Inspecting $dir_name\n" );

    my @plain_files = ();
    my @dirs  = ( $dir_name );
    my @links = ();

    my $path;
    while( scalar @dirs > 0 ) {
	$path = $dirs[ 0 ];
	if( ! &get_directory_content( $path,
				      \@links, \@dirs, \@plain_files ) ) {
	    die "Unable to find resolution of data type $datatype\n";
	}
    }
    &info( "Found path $path\n" );
    if( $path =~ m#\d\d\d\d.1?\d$# ) {
	$resolution = 604800;
    } elsif( $path =~ m#\d\d\d\d.1?[0-9].[123]?\d$# ) {
	$resolution = 86400;
    } elsif( $path =~ m#\D(60|300|900|1800|3600)$# ) {
	$resolution = $1;
    } else {
	die "Unable to find resolution.\n";
    }

    &info( "Resolution $resolution\n" );
    return;
}


sub file_size_of_ok_file {
    my $file = shift @_;
    &info( "Checking file size of $file\n" );
    if( -z $file ) {
	&info( "$file is zero size\n" );
	return 0;
    }
    if( OpeniT::Contrib::File::is_gz_file( $file ) ) {
	my $size = OpeniT::Contrib::File::gz_file_size( $file );
	if( ! defined $size ) {
	    &info( "$file seems to be corrupt or non-gz\n" );
	    return 0;
	}
	&info( "Gzipped file size: ${size}B\n" );
	return $size;
    }
    my $size = -s $file;
    &info( "Plain file size: ${size}B\n" );
    return $size;
}

sub remove_files {
    my( @files ) = @_;
    &info( "Removing file(s):\n" . join( "\n", @files ) . "\n" );
    my $ok = 1;
    foreach my $file( @files ) {
	if( ! unlink $file ) {
	    warn "Could not unlink $file: $!\n";
	    $ok = 0;
	}
    }
    return $ok;
}


sub process_data {
    my( $dt ) = @_;
    $datatype = $dt;
    &info( "Data type path: $db_path$filesep${datatype}\n" );
    if( ! -d "$db_path$filesep$datatype${filesep}W" ||
	! -d "$db_path$filesep$datatype${filesep}M" ) {
	print "Missing W or M directory for data type $dt, can not recover.\n";
	return;
    }
    &check_propagate();
    &find_resolution( "$db_path$filesep$datatype${filesep}M" );
    &process_tree( "$db_path$filesep$datatype" );
    &check_links( "$db_path$filesep$datatype" );
}


sub get_directory_content {
    my( $dir_name, $links, $dirs, $plain_files ) = @_;

    &info( "Getting directory content $dir_name\n" );

    my @plain_files = ();
    my @dirs  = ();
    my @links = ();

    # Remove trailing /
    $dir_name =~ s#/+$##;
    $dir_name =~ s#\\+$##;

    # Find links, dirs and plain files
    my $dir_handle = new DirHandle $dir_name;
    my $dir_entry;
    if (! defined($dir_handle)) {
	warn "Couldn't read dir $dir_name: $!\n";
	return 0;
    }
    while (defined($dir_entry = $dir_handle->read)) {
	next if ($dir_entry eq "." || $dir_entry eq "..");
	my $path = "$dir_name$filesep$dir_entry";
	if( -l $path ) {
	    &info( "$path is a link\n" );
	    push @links, $path;
	} elsif(-d $path ) {
	    &info( "$path is a directory\n" );
	    push @dirs, $path;
	} elsif( -f $path ) {
	    &info( "$path is a plain file\n" );
	    push @plain_files, $path;
	} else {
	    warn "Unknown file type $path. Skipped. Expect other warnings.\n";
	}
    }
    undef $dir_handle;

    &info( "Links:\n" . join( "\n", @links ) . "\n" );
    &info( "Directories:\n" . join( "\n", @dirs ) . "\n" );
    &info( "Plain files:\n" . join( "\n", @plain_files ) . "\n" );

    @$links = @links;
    @$dirs = @dirs;
    @$plain_files = @plain_files;
    return 1;
}

sub find_temp {
    my( @files ) = @_;
    &info( "Finding data-*-*-*.temp files\n" );
    my @temp = ();
    foreach my $file( @files ) {
	if( $file =~ m/data-\d+-\d+-\d+.temp$/ ) {
	    push @temp, $file;
	}
	if( $file =~ m/data-\d+-\d+-\d+.temp.tmp$/ ) {
	    &remove_files( $file );
	}
    }
    &info( join( "\n", @temp ) . "\n" );
    return @temp;
}

sub find_basic {
    my( @files ) = @_;
    &info( "Finding data.* (non-temp) files\n" );
    my @basic = ();
    foreach my $file( @files ) {
	if( $file =~ m/data(\.(cpy|temp|sumbak|tmp|gz))*$/ ) {
	    push @basic, $file;
	}
    }
    &info( join( "\n", @basic ) . "\n" );
    return @basic;
}

sub process_propagate_directory {
    my @plain_files = @_;

    &info( "Processing propagate leaf\n" );

    my @temp = &find_temp( @plain_files );
    if( scalar @temp ) {
	&keep_newest_temp( @plain_files );
	return KEEP_DIR;
    }

    # Only data.* (not data-*-*-*.temp)
    my @basic = &find_basic( @plain_files );
    if( scalar @basic == 0 ) {
	&info( "No basic file\n" );
	return REMOVE_DIR;
    }
    if( scalar @basic == 1 && $basic[ 0 ] =~ m/data(\.gz)*$/ ) {
	return KEEP_DIR;
    }

    if( &keep_oldest_basic( @basic ) == REMOVE_DIR ) {
	return REMOVE_DIR;
    }
    return KEEP_DIR;
}

sub compare_data_with_data_tmp {
    my( $data, $data_tmp, $files ) = @_;
    if( -f $data_tmp && &file_size_of_ok_file( $data_tmp ) == 0 ) {
	&info( "Removing empty $data_tmp\n" );
	if( &remove_files( $data_tmp ) ){
	    delete $$files{ $data_tmp };
	}
    }
    if( -f $data && &file_size_of_ok_file( $data ) == 0 ) {
	&info( "Removing empty $data\n" );
	if( &remove_files( $data ) ) {
	    delete $$files{ $data };
	}
    }
    if( -f $data_tmp && -f $data && -M $data_tmp > -M $data ) {
	if( &remove_files( $data ) ) {
	    delete $$files{ $data };
	}
	if( ! rename $data_tmp, $data ) {
	    die "FATAL: Could not rename $data_tmp: $!\n";
	}
	&info( "Renamed $data_tmp to $data\n" );
    }
    if( -f $data_tmp ) {
	if( &remove_files( $data_tmp ) ){
	    delete $$files{ $data_tmp };
	}
    }
}

sub remove_all_except {
    my( $except, $files ) = @_;
    foreach my $file( keys %$files ) {
	&info( "Checking for removal: $file\n" );
	my $basename = &File::Basename::basename( $file );
	&info( "basename $basename\n" );
	if( $basename ne $except ) {
	    if( &remove_files( $file ) ) {
		delete $$files{ $file };
	    }
	}
    }
}

sub detect_file_base {
    my( $dir ) = @_;
    my $found = 0;
    my( $tmp, $gzipped );
    if( -f "$dir${filesep}data.temp" || -f "$dir${filesep}data.gz.temp" ) {
	$tmp = ".temp"; # This will happen after a --redo crash
    } else {
	$tmp = ".cpy";
    }
    &info( "Selecting $tmp\n" );
    my( $data, $data_tmp );
    my @gzips = ( "", ".gz" );
    foreach my $gzip( @gzips ) {
	$data = "$dir${filesep}data$gzip";
	$data_tmp = "$dir${filesep}data$gzip$tmp";
	&info( "Testing $data and $data_tmp\n" );
	if( -f $data || -f $data_tmp ) {
	    &info( "Gzip is '$gzip'\n" );
	    if( $gzip ) {
		$gzipped = ".gz";
	    } else {
		$gzipped = "";
	    }
	    $found = 1;
	    last;
	}
    }
    if( ! $found ) {
	die "Can not clean up directory $dir, because of content.\n";
    }
    return( $data, $data_tmp );
}

sub process_add_directory {
    my( $dir, @plain_files ) = @_;

    &info( "Processing add directory\n" );

    my @basic = &find_basic( @plain_files );
    my @temp = &find_temp( @plain_files );
    if( scalar @basic == 0 && scalar @temp == 0 ) {
	return REMOVE_DIR;
    }
    my %files = ();
    foreach my $file( @basic ) {
	$files{ $file } = 1;
    }
    my( $data, $data_tmp );
    if( scalar keys %files ) {
	( $data, $data_tmp ) = &detect_file_base( $dir );
	&info( "Working on $data and $data_tmp\n" );
	&compare_data_with_data_tmp( $data, $data_tmp, \%files );
	my $base = File::Basename::basename( $data ); # data or data.gz
	&remove_all_except( $base, \%files );
    }
    if( scalar %files || scalar @temp > 0 ) {
	return KEEP_DIR;
    }
    return REMOVE_DIR;
}

sub order_by_age {
    my( @files ) = @_;
    &info( "Ordering files by age\n" );
    foreach my $file( @files ) {
	my $age = -M $file; # Higher number = older file
	$file = [ $file, $age ];
    }
    @files = sort { $$a[ 1 ] <=> $$b[ 1 ] } @files; # newest first
    foreach my $file( @files ) {
	&info( "Age: $$file[ 1 ]. File $$file[ 1 ]\n" );
	$file = $$file[ 0 ];
    }
    return @files;
}

sub keep_oldest_basic {
    my( @plain_files ) = @_;
    my @basic = &find_basic( @plain_files );
    &info( "Keeping oldest file\n" );
    @basic = &order_by_age( @basic ); # newest first
    my $wanted = pop @basic;
    while( scalar @basic && file_size_of_ok_file( $wanted ) == 0 ) {
	&remove_files( $wanted );
	$wanted = pop @basic;
    }
    &remove_files( @basic );

    if( file_size_of_ok_file( $wanted ) == 0 ) {
	&remove_files( $wanted );
	return REMOVE_DIR;
    }

    my $new_name = &OpeniT::Contrib::File::get_temp_file_name( $wanted );
    &info( "Creating temp file: $new_name\n" );
    if( &OpeniT::Contrib::File::is_gz_file( $wanted ) ) {
	my $gunzipped = &OpeniT::Contrib::File::gunzip_file( $wanted );
	if( ! $gunzipped ) {
	    warn "Deleting corrupt gz-file: $wanted\n";
	    &remove_files( $wanted );
	    return REMOVE_DIR;
	}
	$wanted = $gunzipped;
    }
    &info( "Keeping $wanted\n" );
    if( ! rename( $wanted, $new_name ) ) {
	warn "Could not rename $wanted to $new_name: $!\n";
	return KEEP_DIR;
    }
    &info( "Called $new_name\n" );
    return KEEP_DIR;

}

sub keep_newest_temp {
    my( @plain_files ) = @_;

    &info( "Keeping newest temp file\n" );
    my @basic = &find_basic( @plain_files );
    &remove_files( @basic );

    my @temp = &find_temp( @plain_files );
    @temp = order_by_age( @temp ); # newest first
    my $wanted = shift @temp;

    # We assume never two temp files with same age
    &remove_files( @temp );
    &info( "Keeping $wanted\n" );
}

sub check_links {
    my( $dir_name ) = @_;

    my @plain_files = ();
    my @dirs  = ();
    my @links = ();

    &info( "processing tree with respect to links: $dir_name\n" );

    if( ! get_directory_content( $dir_name,
				 \@links, \@dirs, \@plain_files ) ) {
	warn "Unable to process $dir_name\n";
	return KEEP_DIR;
    }

    my $removed_dirs = 0;
    foreach my $dir( @dirs ) {
	&info( "Recursive on $dir\n" );
	if( &check_links( $dir ) == REMOVE_DIR ) {
	    if( ! rmdir $dir ) {
		warn "Failed to remove dir $dir: $!\n";
	    } else {
		$removed_dirs++;
	    }
	}
    }

    my $removed_links = 0;
    foreach my $link( @links ) {
	my $path = "$dir_name$filesep" . readlink $link;
	if( -d "$path" || $link !~ m/lock$/ ) {
	    next;
	}
	&info( "Removing stray link $link\n" );
	if( &remove_files( $link ) ) {
	    $removed_links++;
	}
    }
    if( scalar @dirs == $removed_dirs &&
	scalar @links == $removed_links &&
	scalar @plain_files == 0 ) {
	return REMOVE_DIR;
    }
    if( $resolution != 604800 &&
	scalar @dirs == $removed_dirs &&
	scalar @plain_files == 1 &&
	$plain_files[ 0 ] !~ m/data(\.gz|-\d+-\d+-\d+\.temp)?$/ &&
	scalar @links == $removed_links ) {
	if( &remove_files( $plain_files[ 0 ] ) ) {
	    return REMOVE_DIR;
	}
    }
    return KEEP_DIR;
}

sub process_tree {
    my( $dir_name ) = @_;

    my @plain_files = ();
    my @dirs  = ();
    my @links = ();

    &info( "processing tree: $dir_name\n" );

    if( ! get_directory_content( $dir_name,
				 \@links, \@dirs, \@plain_files ) ) {
	warn "Unable to process $dir_name\n";
	return KEEP_DIR;
    }

    for( my $idx = scalar @plain_files; $idx > 0; $idx-- ) {
	my $file = $plain_files[ $idx - 1 ];
	if( $file =~ m/data-\d+-\d+-\d+.temp$/ ) {
	    next;
	}
	if( -z $file ||
	    ( -s $file < 100 &&
	      &OpeniT::Contrib::File::is_gz_file( $file ) &&
	      &OpeniT::Contrib::File::gz_file_size( $file ) == 0 ) ) {
	    if( &remove_files( $file ) ) {
		splice @plain_files, $idx - 1, 1, ();
	    }
	}
    }

    if( scalar @links == 0 && scalar @dirs == 0 && scalar @plain_files == 0 ) {
	# Empty directory
	&info( "Empty directory $dir_name\n" );
	if( $resolution != 604800 ) {
	    &info( "Removing $dir_name\n" );
	    return REMOVE_DIR;
	}
	&info( "Keep it, it seems to not be corrupt, but empty from start\n" );
	return KEEP_DIR;
    }

    my $removed_dirs = 0;
    foreach my $dir( @dirs ) {
	&info( "Recursive on $dir\n" );
	if( &process_tree( $dir ) == REMOVE_DIR ) {
	    if( ! rmdir $dir ) {
		warn "Failed to remove dir $dir: $!\n";
	    } else {
		$removed_dirs++;
	    }
	}
    }

    if( $is_propagate ) {
	my $leaf = 0;
	if( &File::Basename::basename( $dir_name ) eq $resolution ) {
	    $leaf = 1;
	}
	if( $resolution == 86400 && $dir_name =~ m/M\D\d\d\d\d\D\d\d?\D\d\d?$/ ) {
	    $leaf = 1;
	}
	if( $resolution == 604800 && $dir_name =~ m/W\D\d\d\d\d\D\d\d?$/ ) {
	    $leaf = 1;
	}
	if( $leaf ) {
	    my $status = &process_propagate_directory( @plain_files );
	    if( scalar @links == 0 && scalar @dirs == $removed_dirs ) {
		return $status;
	    }
	    return KEEP_DIR;
	}
	my $status = &process_add_directory( $dir_name, @plain_files );
	if( scalar @links == 0 && scalar @dirs == $removed_dirs ) {
	    return $status;
	}
	return KEEP_DIR;
    }
    my $status = &process_add_directory( $dir_name, @plain_files );
    if( scalar @links == 0 && scalar @dirs == $removed_dirs ) {
	return $status;
    }
    return KEEP_DIR;
}

sub info {
    my $string = shift @_;
    if( $info ) {
	print $string;
    }
}

1;
