#
#
# $Id: File.pm,v 1.1.2.3 2012/05/08 13:12:21 kyri Exp $
#
# (C) Copyright 2012 by Open iT inc., All rights reserved.
#
# This software may be used or copied only with the written permission by
# Open Systems Consultans a.s or in accordance with the terms and conditions
# stipulated in the agreement under which this software has been supplied.
#
# This file is part of Open iT.
#
#HEAD##########################################################################
#
# @NAME:
#   File
# @AUTHOR:
#   Kjetil Yri
# @CREATED:
#   Thu Apr 12 09:55:28 MET 2012
# @DESCRIPTION:
#   Module for file operations.
#
#/HEAD#########################################################################

use strict;

package OpeniT::Contrib::File;

use FileHandle;

use OpeniT::Contrib::Debug qw ( debug );
use Compress::Zlib qw(/gz/ /Z_/);

require Exporter;

@OpeniT::Contrib::DatabaseExtraction::ISA    = qw( Exporter );
@OpeniT::Contrib::DatabaseExtraction::EXPORT =
  qw( file_separator
      is_gz_file
      gz_file_size
      gunzip_file
    );

@OpeniT::Contrib::DatabaseExtraction::EXPORT_OK = qw();

my $filesep;
if( $^O eq "MSWin32" ) {
    $filesep = "\\";
} else {
    $filesep = "/";
}

sub file_separator {
    return $filesep;
}

####
# Input: zipped file, of any other file
# Output: 1 if file is zipped and has correct format
#         0 if not
#         undef on error
#
sub is_gz_file {
    my $filename = shift @_;
    my $stat = open( GZFILE, "<$filename" );
    if( ! $stat ) {
	debug( "$filename is not gzipped" );
	return 0;
    }
    my $buf;
    if( read( GZFILE, $buf, 2 ) != 2 ) {
	close( GZFILE );
	debug( "$filename is not gzipped" );
	return 0;
    }
    if( ord substr( $buf, 0, 1 ) != 31 || ord substr( $buf, 1, 1 ) != 139 ) {
	close( GZFILE );
	debug( "$filename is not gzipped" );
	return 0;
    }
    close( GZFILE );
    debug( "$filename is gzipped" );
    return 1;
}

####
# Input: zipped file
# Output: uncompressed size
#         undef on error
#
sub gz_file_size {
    my $filename = shift @_;

    my $stat = open( GZFILE, "<$filename" );
    if( ! $stat ) {
	ERROR( "Could not find file size of: $filename: $!",undef, "noexit" );
	return undef;
    }
    my $buf;
    if( read( GZFILE, $buf, 2 ) != 2 ) {
	ERROR( "Could not read for finding size of $filename: $!", undef,
	       "noexit" );
	close( GZFILE );
	return undef;
    }
    if( ord substr( $buf, 0, 1 ) != 31 || ord substr( $buf, 1, 1 ) != 139 ) {
	ERROR( "Could not find gz size of non-gz file: $filename", undef,
	       "noexit" );
	close( GZFILE );
	return undef;
    }
    if( ! seek( GZFILE, -4, 2 ) ) {
	ERROR( "Could not seek for size of gz file $filename: $!",
	       undef, "noexit" );
	close( GZFILE );
	return undef;
    }
    if( read( GZFILE, $buf, 4 ) != 4 ) {
	ERROR( "Could not read for finding size of $filename: $!", undef,
	       "noexit" );
	close( GZFILE );
	return undef;
    }
    close( GZFILE );
    my $size = unpack( "V", $buf );
    return $size;
}

####
# Input: filename of zipped file
# Output: Output file name if successful gunzip
#         0 if error
# Note: If file extension is ".gz" the file will be renamed
#       If ".gz" extension is used, "force" must be used if orig file exists
sub gunzip_file {
    my( $filename, $force ) = @_;
    debug( "gunzip_file( $filename )" );
    if( ! -f $filename ) {
	ERROR( "No such file: $filename", undef, "noexit" );
	return 0;
    }
    if( ! &is_gz_file( $filename ) ) {
	ERROR( "Cannot gunzip non-gzip file: $filename", undef, "noexit" );
	return 0;
    }
    my $outfilename = $filename;
    if( $outfilename =~ s/\.gz$// ) {
	if( -f $outfilename && ! $force ) {
	    ERROR( "gunzip: Can not overwrite $outfilename", undef, "noexit" );
	    return 0;
	}
    }
    my $gztmp = ".gz-tmp";
    my $infile = gzopen( $filename, "r" );
    if( ! defined $infile ) {
	ERROR( "gunzip: Could not open $filename: $gzerrno", undef, "noexit" );
	return 0;
    }
    my $outfile = new FileHandle;
    if( ! open( $outfile, ">$outfilename$gztmp" ) ) {
	$infile->gzclose();
	ERROR( "gunzip: Could not open $outfilename$gztmp: $!",
	       undef, "noexit" );
	return 0;
    }
    my $line;
    while( $infile->gzreadline( $line ) ) {
	if( ! print $outfile $line ) {
	    ERROR( "gunzip: Could not write to $outfilename$gztmp: $!",
		   undef, "noexit" );
	    $infile->gzclose();
	    close( $outfile );
	    if( ! unlink "$outfilename$gztmp" ) {
		debug( "Could not unlink $outfilename$gztmp: $!" );
	    }
	    return 0;
	}
    }
    if( ! $infile->gzeof() ) {
	ERROR( "gunzip: Could not read $filename: $gzerrno", undef, "noexit" );
	$infile->gzclose();
	close( $outfile );
	if( ! unlink "$outfilename$gztmp" ) {
	    debug( "Could not unlink $outfilename$gztmp: $!" );
	}
	return 0;
    }
    $infile->gzclose();
    close( $outfile );
    if( ! rename "$outfilename$gztmp", $outfilename ) {
	ERROR( "Could not rename $outfilename$gztmp to $outfilename",
	       undef, "noexit" );
	return 0;
    }
    debug( "Successfully gunzipped $filename, output $outfilename" );
    if( $filename ne $outfilename ) {
	debug( "Removing $filename" );
	if( ! unlink $filename ) {
	    ERROR( "Could not unlink $filename: $!", undef, "noexit" );
	}
    }
    return $outfilename;
}

sub get_temp_file_name {
    my( $old ) = @_;
    my $dir;
    if( ! -d $old ) {
	$old = &File::Basename::dirname( $old );
    }
    my $new_name = "$old${filesep}data-" . time() . "-$$-1.temp";
    while( -f $new_name ) {
	$new_name =~ m/data-\d+-\d+-(\d+).temp/;
	my $count = $1 + 1;
	$new_name =~ s/(data-\d+-\d+-)\d+.temp/$1$count.temp/;
    }
    return $new_name;
}



1;
